import { describe, it, expect, beforeAll } from "@jest/globals";
import { JSDOM } from "jsdom";
import SigningService from "../src/services/SigningService";
import { EMAIL_VENDORS } from "../src/services/StatusesService";
import {
  arrayBufferToBase64,
  CryptoService,
  AttachmentsSignatures,
  AttachmentData,
} from "../src";
import { SEVERITY_LEVELS } from "../src/services/SigningService/severityLevels";

describe("SigningService tests", () => {
  let signingService: SigningService;
  let publicKeyPEM;
  let privateKeyPEM;

  beforeAll(async () => {
    const keys = await CryptoService.generateRSAKeys();
    publicKeyPEM = keys.publicKeyPEM;
    privateKeyPEM = keys.privateKeyPEM;

    signingService = new SigningService();

    /**
     * Must be defined by integrators.
     * @param value
     */
    signingService.helpers.signRSAExternal = async (value: string) => {
      return arrayBufferToBase64(
        await CryptoService.signRSA(
          privateKeyPEM,
          new TextEncoder().encode(value)
        )
      );
    };

    /**
     * Can be amended by integrator.
     * Test is using JSDOM library.
     */
    signingService.helpers.parseHtml = (htmlString) => {
      const jsdom = new JSDOM(htmlString);
      return jsdom.window.document;
    };
  });

  /**
   * https://code.vereign.com/light/documentation/-/blob/master/Validation.md#components-mode
   */
  describe("Components verification mode", () => {
    it("Checks a valid test case", async () => {
      const senderHtml = "<div><p>Beep Bop</p><span>Beep</span></div>";
      const senderPlain = "Beep Bop Beep";

      const attachmentsData: AttachmentData[] = [
        {
          filename: "attachment1",
          type: "raw",
          content: "EPYTHV1oJzOpstj7lv3DT8KGMFqr58",
        },
        {
          filename: "attachment2",
          type: "raw",
          content: "kQrs1M1Qy73PCfq4c3NDo6GdxV3cbaUGyIvVDV",
        },
      ];

      /**
       * Signing logic
       */
      const {
        partsSignatures,
        rabinFingerprints,
      } = await signingService.signMimeComponents(
        EMAIL_VENDORS.GMAIL,
        senderHtml,
        senderPlain,
        attachmentsData
      );

      /**
       * Validation logic
       */
      const receiverHtml = "<div><p>Beep Bop</p><span>Beep</span></div>";
      const receiverPlain = "Beep Bop Beep";
      const receiverAttachments = [...attachmentsData];

      const validationResults = await signingService.verifyMimeComponents(
        publicKeyPEM,
        EMAIL_VENDORS.GMAIL,
        receiverHtml,
        receiverPlain,
        receiverAttachments,
        partsSignatures,
        rabinFingerprints
      );

      expect(validationResults.htmlPartsSignatureValid).toBeTruthy();
      expect(validationResults.plainPartsSignatureValid).toBeTruthy();
      expect(validationResults.pseudoPlainPartsSignatureValid).toBeTruthy();
      expect(validationResults.attachmentsPartsSignatureValid).toBeTruthy();
      expect(validationResults.severityLevel).toEqual(SEVERITY_LEVELS.VALID);
    });

    it("Signs/verifies empty values", async () => {
      /**
       * Signing logic
       */
      const {
        partsSignatures,
        rabinFingerprints,
      } = await signingService.signMimeComponents(
        EMAIL_VENDORS.GMAIL,
        "",
        "",
        []
      );

      /**
       * Validation logic
       */

      const validationResults = await signingService.verifyMimeComponents(
        publicKeyPEM,
        EMAIL_VENDORS.GMAIL,
        "",
        "",
        [],
        partsSignatures,
        rabinFingerprints
      );

      expect(validationResults.htmlPartsSignatureValid).toBeTruthy();
      expect(validationResults.plainPartsSignatureValid).toBeTruthy();
      expect(validationResults.pseudoPlainPartsSignatureValid).toBeTruthy();
      expect(validationResults.attachmentsPartsSignatureValid).toBeTruthy();
      expect(validationResults.severityLevel).toEqual(SEVERITY_LEVELS.VALID);
    });
  });

  describe("Attachments signing/verification", () => {
    it("Checks a valid test case", async () => {
      const fileContentBytes = "abacadaba";
      const fileName = "abacadaba.txt";

      const hash = await signingService.helpers.cryptoService.MD5(
        fileContentBytes
      );

      const attachmentData: AttachmentData = {
        filename: fileName,
        type: "raw",
        content: fileContentBytes,
      };

      const attachmentHash: AttachmentData = {
        filename: fileName,
        type: "hash",
        content: arrayBufferToBase64(hash),
        hashingAlgorithm: "md5",
      };

      const attachments = [];
      attachments.push(attachmentData);
      attachments.push(attachmentHash);

      const signedAttachmentsData = await signingService.signAttachments(
        attachments
      );

      const signatures: AttachmentsSignatures = Object.keys(
        signedAttachmentsData
      ).reduce((acc, hash) => {
        const attachmentsData = signedAttachmentsData[hash];
        return Object.assign(acc, {
          [hash]: attachmentsData.map((attachmentData) => ({
            signature: attachmentData.signature,
            hashingAlgorithm: attachmentData.hashingAlgorithm,
          })),
        });
      }, {});

      let attachmentsToVerify = [];

      attachmentsToVerify.push(attachmentData);

      let verificationResult = await signingService.verifyAttachments(
        publicKeyPEM,
        attachmentsToVerify,
        signatures
      );

      expect(verificationResult.every((verified) => !!verified)).toBeTruthy();

      const attachmentData1 = {
        filename: fileName,
        type: "raw",
        content: "test",
      };

      attachmentsToVerify = [];
      attachmentsToVerify.push(attachmentData1);

      verificationResult = await signingService.verifyAttachments(
        publicKeyPEM,
        attachmentsToVerify,
        signatures
      );

      expect(verificationResult.every((verified) => !!verified)).toBeFalsy();

      const attachmentData2 = {
        filename: "test",
        type: "raw",
        content: "test",
      };

      attachmentsToVerify = [];

      attachmentsToVerify.push(attachmentData);
      attachmentsToVerify.push(attachmentData1);
      attachmentsToVerify.push(attachmentData2);
      attachmentsToVerify.push(attachmentHash);

      verificationResult = await signingService.verifyAttachments(
        publicKeyPEM,
        attachmentsToVerify,
        signatures
      );

      expect(verificationResult[0]).toBeTruthy();
      expect(verificationResult[1]).toBeFalsy();
      expect(verificationResult[2]).toBeFalsy();
      expect(verificationResult[3]).toBeTruthy();
    });
  });
});
