"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const mime_normalizer_1 = require("@vereign/mime-normalizer");
const rka_1 = require("../../utils/rka");
const index_1 = require("../../index");
const severityLevels_1 = require("./severityLevels");
class SigningService {
    constructor() {
        this.cryptoService = index_1.CryptoService;
        this.helpers = {
            signRSAExternal: () => __awaiter(this, void 0, void 0, function* () {
                throw new Error("Please provide external function to calculate RSA signature");
            }),
            parseHtml: (htmlString) => {
                const parser = new DOMParser();
                return parser.parseFromString(htmlString, "text/html");
            },
            cryptoService: index_1.CryptoService,
            arrayBufferToBase64: index_1.arrayBufferToBase64,
        };
    }
    /**
     * Function implementing COMPONENTS signing routine
     * https://code.vereign.com/light/documentation/-/blob/master/Validation.md#components-mode
     * @param vendor - email vendor (GMAIL | OUTLOOK | ROUNDCUBE | GENERIC_MIME)
     * @param html - MIME html part
     * @param plain - MIME plain part
     * @param attachments - an array of attachments hashes
     */
    signMimeComponents(vendor, html, plain, attachments) {
        return __awaiter(this, void 0, void 0, function* () {
            const document = this.helpers.parseHtml(html);
            const normalizedHTML = mime_normalizer_1.HTMLNormalizer.normalizeVendorHtml(document, vendor);
            const normalizedPlain = mime_normalizer_1.PlainNormalizer.normalizePlain(plain);
            const pseudoPlainPart = mime_normalizer_1.HTMLNormalizer.extractPseudoPlainPart(document);
            const hashedAttachments = yield this.hashAttachments(attachments, false);
            const [htmlPartsSignature, plainPartsSignature, pseudoPlainPartsSignature, attachmentsPartsSignature,] = yield this.getPartsSignatures([
                [normalizedHTML],
                [normalizedPlain],
                [pseudoPlainPart],
                hashedAttachments.map((attachment) => attachment.content),
            ]);
            const rabinFingerprints = {
                htmlPartRabinFingerprint: normalizedHTML
                    ? rka_1.default.getRabinFingerprint(normalizedHTML)
                    : "",
                plainPartRabinFingerprint: normalizedPlain
                    ? rka_1.default.getRabinFingerprint(normalizedPlain)
                    : "",
                pseudoPlainPartRabinFingerprint: pseudoPlainPart
                    ? rka_1.default.getRabinFingerprint(pseudoPlainPart)
                    : "",
                htmlPartSize: +(normalizedHTML && normalizedHTML.length),
                plainPartSize: +(normalizedHTML && normalizedPlain.length),
                pseudoPlainPartSize: +(pseudoPlainPart && pseudoPlainPart.length),
            };
            return {
                partsSignatures: {
                    htmlPartsSignature,
                    plainPartsSignature,
                    pseudoPlainPartsSignature,
                    attachmentsPartsSignature,
                },
                rabinFingerprints,
            };
        });
    }
    getPartsSignatures(partsBatch) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield Promise.all(partsBatch.map((parts) => __awaiter(this, void 0, void 0, function* () {
                const partsHash = yield this.preparePartsForSigning(parts);
                if (partsHash) {
                    return yield this.helpers.signRSAExternal(partsHash);
                }
                else {
                    return "";
                }
            })));
        });
    }
    /**
     * Function implementing COMPONENTS signing routine
     * https://code.vereign.com/light/documentation/-/blob/master/Validation.md#components-mode
     * @param publicKeyPEM
     * @param vendor - email vendor (GMAIL | OUTLOOK | ROUNDCUBE | GENERIC_MIME)
     * @param html - MIME html part
     * @param plain - MIME plain part
     * @param attachments - an array of attachments hashes
     * @param partsSignatures
     * @param partsRabinFingerprints
     */
    verifyMimeComponents(publicKeyPEM, vendor, html, plain, attachments, partsSignatures, partsRabinFingerprints) {
        return __awaiter(this, void 0, void 0, function* () {
            const document = this.helpers.parseHtml(html);
            const normalizedHTML = mime_normalizer_1.HTMLNormalizer.normalizeVendorHtml(document, vendor);
            const normalizedPlain = mime_normalizer_1.PlainNormalizer.normalizePlain(plain);
            const pseudoPlainPart = mime_normalizer_1.HTMLNormalizer.extractPseudoPlainPart(document);
            const hashedAttachments = yield this.hashAttachments(attachments, false);
            const originalNormalizedHtml = rka_1.default.findRKASubstring(normalizedHTML, partsRabinFingerprints.htmlPartRabinFingerprint, partsRabinFingerprints.htmlPartSize);
            const originalNormalizedPlain = rka_1.default.findRKASubstring(normalizedPlain, partsRabinFingerprints.plainPartRabinFingerprint, partsRabinFingerprints.plainPartSize);
            const originalPseudoPlain = rka_1.default.findRKASubstring(pseudoPlainPart, partsRabinFingerprints.pseudoPlainPartRabinFingerprint, partsRabinFingerprints.pseudoPlainPartSize);
            const [htmlPartsValid, plainPartsValid, pseudoPlainPartsValid, attachmentsPartsValid,] = yield this.verifyPartsSignatures(publicKeyPEM, [
                {
                    signature: partsSignatures.htmlPartsSignature,
                    parts: [originalNormalizedHtml],
                },
                {
                    signature: partsSignatures.plainPartsSignature,
                    parts: [originalNormalizedPlain],
                },
                {
                    signature: partsSignatures.pseudoPlainPartsSignature,
                    parts: [originalPseudoPlain],
                },
                {
                    signature: partsSignatures.attachmentsPartsSignature,
                    parts: hashedAttachments.map((attachment) => attachment.content),
                },
            ]);
            const severityStateIndex = severityLevels_1.findSeverityStateIndex(htmlPartsValid, plainPartsValid, pseudoPlainPartsValid, attachmentsPartsValid);
            return {
                htmlPartsSignatureValid: htmlPartsValid,
                plainPartsSignatureValid: plainPartsValid,
                pseudoPlainPartsSignatureValid: pseudoPlainPartsValid,
                attachmentsPartsSignatureValid: attachmentsPartsValid,
                severityLevel: severityLevels_1.SEVERITY_LEVEL_PER_STATE[severityStateIndex],
                statusMessage: severityLevels_1.SEVERITY_MESSAGE_PER_STATE[severityStateIndex],
            };
        });
    }
    verifyPartsSignatures(publicKeyPEM, partsBatch) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield Promise.all(partsBatch.map(({ signature, parts }) => __awaiter(this, void 0, void 0, function* () {
                const partsHash = yield this.preparePartsForSigning(parts);
                if (partsHash) {
                    return yield this.helpers.cryptoService.verifyRSASignature(publicKeyPEM, Buffer.from(partsHash), Buffer.from(signature, "base64"));
                }
                else {
                    // If there's no hash, but signature exists, then it's invalid
                    return !signature;
                }
            })));
        });
    }
    /**
     * Function implementing specification
     * https://code.vereign.com/light/documentation/-/blob/master/Validation.md#partssignaturealgorithm
     *
     * @param parts - array of strings representing parts
     */
    preparePartsForSigning(parts) {
        return __awaiter(this, void 0, void 0, function* () {
            const hashedParts = yield Promise.all(parts.map((part) => __awaiter(this, void 0, void 0, function* () {
                return yield this.helpers.cryptoService.SHA256(part);
            })));
            return hashedParts
                .map((hash) => this.helpers.arrayBufferToBase64(hash))
                .sort()
                .join("\n");
        });
    }
    hashAttachments(attachments, useHashingAlgorithm) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = [];
            const cryptoService = this.helpers.cryptoService;
            const hashingFunctions = {
                sha1: cryptoService.SHA1,
                sha256: cryptoService.SHA256,
                sha384: cryptoService.SHA384,
                sha512: cryptoService.SHA512,
                md5: cryptoService.MD5,
            };
            for (let i = 0; i < attachments.length; i++) {
                const attachment = attachments[i];
                if (attachment.type == "raw") {
                    let hash;
                    let hashingAlgorithm;
                    if (useHashingAlgorithm == true) {
                        hashingAlgorithm = attachment.hashingAlgorithm;
                        const hashingFunction = hashingFunctions[hashingAlgorithm];
                        if (hashingFunction) {
                            hash = yield hashingFunction(attachment.content);
                        }
                        else {
                            throw new Error(`No such hashing algorithm supported: "${hashingAlgorithm}"`);
                        }
                    }
                    else {
                        hashingAlgorithm = "sha256";
                        hash = yield this.helpers.cryptoService.SHA256(attachment.content);
                    }
                    const attachmentHashed = Object.assign(Object.assign({}, attachment), { content: hash, type: "hash", hashingAlgorithm: hashingAlgorithm });
                    result.push(attachmentHashed);
                }
                else {
                    result.push(attachment);
                }
            }
            return result;
        });
    }
    /**
     * @param attachments - an array of attachments data objects
     */
    signAttachments(attachments) {
        return __awaiter(this, void 0, void 0, function* () {
            const hashedAttachments = yield this.hashAttachments(attachments, false);
            const result = {};
            for (let i = 0; i < hashedAttachments.length; i++) {
                const hashedAttachment = hashedAttachments[i];
                const wrappingHash = yield this.helpers.cryptoService.SHA256(hashedAttachment.content);
                const filenameHashBuffer = yield this.helpers.cryptoService.SHA256(hashedAttachment.filename);
                const filenameHash = this.helpers.arrayBufferToBase64(filenameHashBuffer);
                const signature = yield this.helpers.signRSAExternal(this.helpers.arrayBufferToBase64(wrappingHash));
                if (!(filenameHash in result)) {
                    result[filenameHash] = [];
                }
                hashedAttachment.signature = signature;
                result[filenameHash].push(hashedAttachment);
            }
            return result;
        });
    }
    /**
     * @param publicKeyPEM
     * @param attachments - an array of attachments hashes
     * @param signatures
     */
    verifyAttachments(publicKeyPEM, attachments, signatures) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = [];
            for (let i = 0; i < attachments.length; i++) {
                const attachment = attachments[i];
                const filename = attachment.filename;
                const filenameHashBuffer = yield this.helpers.cryptoService.SHA256(filename);
                const filenameHash = this.helpers.arrayBufferToBase64(filenameHashBuffer);
                if (filenameHash in signatures) {
                    const attachmentSignatures = signatures[filenameHash];
                    let attachmentVersionsHashed = [];
                    if (attachment.type == "raw") {
                        const hashingAlgorithms = {};
                        for (let j = 0; j < attachmentSignatures.length; j++) {
                            const signature = attachmentSignatures[j];
                            hashingAlgorithms[signature.hashingAlgorithm] = true;
                        }
                        const attachmentVersions = [];
                        for (const hashingAlgorithm in hashingAlgorithms) {
                            if (hashingAlgorithms[hashingAlgorithm]) {
                                const attachmentVersion = {
                                    filename: filename,
                                    content: attachment.content,
                                    type: "raw",
                                    hashingAlgorithm: hashingAlgorithm,
                                };
                                attachmentVersions.push(attachmentVersion);
                            }
                        }
                        attachmentVersionsHashed = yield this.hashAttachments(attachmentVersions, true);
                    }
                    else {
                        attachmentVersionsHashed.push(attachment);
                    }
                    let valid = false;
                    for (let j = 0; j < attachmentSignatures.length; j++) {
                        if (valid === true) {
                            break;
                        }
                        const signature = attachmentSignatures[j];
                        for (let k = 0; k < attachmentVersionsHashed.length; k++) {
                            const attachmentHash = attachmentVersionsHashed[k];
                            const wrappingHash = yield this.helpers.cryptoService.SHA256(attachmentHash.content);
                            const base64WrappingHash = this.helpers.arrayBufferToBase64(wrappingHash);
                            const isValid = yield this.helpers.cryptoService.verifyRSASignature(publicKeyPEM, Buffer.from(base64WrappingHash), Buffer.from(signature.signature, "base64"));
                            if (isValid == true) {
                                valid = true;
                                break;
                            }
                        }
                    }
                    result.push(valid);
                }
                else {
                    result.push(false);
                }
            }
            return result;
        });
    }
}
exports.default = SigningService;
