// All functions are async in order to be compatible with the ICryptoService API
import md5 from "js-md5";

import * as crypto from "crypto";

import { AESGCMOutput, RSAKeys, ICryptoService } from "./ICryptoService";
import { arrayBufferToBase64 } from "../../utils/common";

const AES_GCM_ALGO = "aes-256-gcm";

const encryptAESGCM = async (data: string): Promise<AESGCMOutput> => {
  const key = crypto.randomBytes(32);
  const iv = crypto.randomBytes(12);
  const cipher = crypto.createCipheriv(AES_GCM_ALGO, key, iv);

  const encrypted = cipher.update(data, "utf8");
  cipher.final();

  const authTag = cipher.getAuthTag();
  const encryptedWithTag = Buffer.concat([
    Buffer.from(encrypted),
    Buffer.from(authTag),
  ]);

  return {
    key: key.buffer,
    iv: iv.buffer,
    data: encryptedWithTag,
  };
};

const decryptAESGCM = async (
  data: ArrayBuffer,
  key: ArrayBuffer,
  iv: ArrayBuffer
): Promise<string> => {
  const decipher = crypto.createDecipheriv(
    AES_GCM_ALGO,
    Buffer.from(key),
    Buffer.from(iv)
  );

  const authTag = data.slice(data.byteLength - 16, data.byteLength);
  const encrypted = data.slice(0, data.byteLength - 16);
  let str = decipher.update(arrayBufferToBase64(encrypted), "base64", "utf8");
  decipher.setAuthTag(Buffer.from(authTag));
  str += decipher.final("utf8");
  return str;
};

const generateRSAKeys = async (): Promise<RSAKeys> => {
  const { privateKey, publicKey } = crypto.generateKeyPairSync("rsa", {
    modulusLength: 4096,
    publicKeyEncoding: {
      type: "spki",
      format: "pem",
    },
    privateKeyEncoding: {
      type: "pkcs8",
      format: "pem",
    },
  });

  return {
    publicKeyPEM: publicKey,
    privateKeyPEM: privateKey,
  };
};

const encryptRSA = async (
  publicKeyPEM: string,
  data: ArrayBuffer
): Promise<ArrayBuffer> => {
  const encryptedData = crypto.publicEncrypt(
    {
      key: publicKeyPEM,
      padding: crypto.constants.RSA_PKCS1_OAEP_PADDING,
      oaepHash: "sha256",
    },
    // We convert the data string to a buffer using `Buffer.from`
    Buffer.from(data)
  );

  return encryptedData.buffer;
};

const decryptRSA = async (
  privateKeyPEM: string,
  data: ArrayBuffer
): Promise<ArrayBuffer> => {
  const encryptedData = crypto.privateDecrypt(
    {
      key: privateKeyPEM,
      padding: crypto.constants.RSA_PKCS1_OAEP_PADDING,
      oaepHash: "sha256",
    },
    // We convert the data string to a buffer using `Buffer.from`
    Buffer.from(data)
  );

  return encryptedData.buffer;
};

const signRSA = async (
  privateKeyPEM: string,
  data: ArrayBuffer
): Promise<ArrayBuffer> => {
  const sign = crypto.createSign("SHA256");
  sign.update(Buffer.from(data));
  sign.end();

  return sign.sign(privateKeyPEM).buffer;
};

const SHA1 = async (
  value: string | ArrayBuffer,
  encoding: BufferEncoding = "utf8"
): Promise<ArrayBuffer> => {
  const bytes = getBytes(value, encoding);
  return crypto.createHash("sha1").update(bytes).digest();
};

const SHA256 = async (
  value: string | ArrayBuffer,
  encoding: BufferEncoding = "utf8"
): Promise<ArrayBuffer> => {
  const bytes = getBytes(value, encoding);
  return crypto.createHash("sha256").update(bytes).digest();
};

const SHA384 = async (
  value: string | ArrayBuffer,
  encoding: BufferEncoding = "utf8"
): Promise<ArrayBuffer> => {
  const bytes = getBytes(value, encoding);
  return crypto.createHash("sha384").update(bytes).digest();
};

const SHA512 = async (
  value: string | ArrayBuffer,
  encoding: BufferEncoding = "utf8"
): Promise<ArrayBuffer> => {
  const bytes = getBytes(value, encoding);
  return crypto.createHash("sha512").update(bytes).digest();
};

const MD5 = async (
  value: string | ArrayBuffer,
  encoding: BufferEncoding = "utf8"
): Promise<ArrayBuffer> => {
  const bytes = getBytes(value, encoding);
  return md5.arrayBuffer(bytes);
};

const getBytes = (
  value: string | ArrayBuffer,
  encoding: BufferEncoding
): Buffer => {
  let bytes;
  if (typeof value === "string") {
    const valueString = value as string;
    bytes = Buffer.from(valueString, encoding);
  } else {
    bytes = Buffer.from(value);
  }
  return bytes;
};

const verifyRSASignature = async (
  publicKeyPEM: string,
  data: ArrayBuffer,
  signature: ArrayBuffer
): Promise<boolean> => {
  const verify = crypto.createVerify("SHA256");
  verify.update(Buffer.from(data));
  verify.end();

  const publicKey = crypto.createPublicKey(Buffer.from(publicKeyPEM, "utf-8"));

  return verify.verify(publicKey, Buffer.from(signature));
};

const implementation: ICryptoService = {
  encryptAESGCM,
  decryptAESGCM,
  generateRSAKeys,
  encryptRSA,
  decryptRSA,
  SHA256,
  signRSA,
  verifyRSASignature,
  SHA512,
  SHA384,
  SHA1,
  MD5,
};

export default implementation;
