import md5 from "js-md5";
import { AESGCMOutput, ICryptoService, RSAKeys } from "./ICryptoService";
import { base64ToArrayBuffer } from "../../utils/common";

async function exportKey(key: CryptoKey): Promise<ArrayBuffer> {
  return crypto.subtle.exportKey("raw", key);
}

const convertPemToBinary = (pem: string) => {
  const lines = pem.split("\n");
  let encoded = "";
  for (let i = 0; i < lines.length; i++) {
    if (
      lines[i].trim().length > 0 &&
      lines[i].indexOf("-BEGIN PRIVATE KEY-") < 0 &&
      lines[i].indexOf("-BEGIN PUBLIC KEY-") < 0 &&
      lines[i].indexOf("-END PRIVATE KEY-") < 0 &&
      lines[i].indexOf("-END PUBLIC KEY-") < 0
    ) {
      encoded += lines[i].trim();
    }
  }
  return base64ToArrayBuffer(encoded);
};

export const verifyRSASignature = async (
  publicKeyPEM: string,
  data: ArrayBuffer,
  signature: ArrayBuffer
): Promise<boolean> => {
  const publicKey = await crypto.subtle.importKey(
    "spki",
    convertPemToBinary(publicKeyPEM),
    {
      name: "RSASSA-PKCS1-v1_5",
      hash: "SHA-256",
    },
    true,
    ["verify"]
  );

  return await crypto.subtle.verify(
    {
      name: "RSASSA-PKCS1-v1_5",
      hash: "SHA-256",
    },
    publicKey,
    signature,
    data
  );
};

const encryptAESGCM = async (data: string): Promise<AESGCMOutput> => {
  const key = await crypto.subtle.generateKey(
    {
      name: "AES-GCM",
      length: 256,
    },
    true,
    ["encrypt", "decrypt"]
  );

  const encoded = new TextEncoder().encode(data);
  const iv = crypto.getRandomValues(new Buffer(12));
  const encrypted = await crypto.subtle.encrypt(
    { name: "AES-GCM", iv: iv },
    key,
    encoded
  );
  return { data: encrypted, key: await exportKey(key), iv };
};

const decryptAESGCM = async (
  data: ArrayBuffer,
  key: ArrayBuffer,
  iv: ArrayBuffer
): Promise<string> => {
  const importedKey = await crypto.subtle.importKey(
    "raw",
    key,
    {
      name: "AES-GCM",
      length: 256,
    },
    true,
    ["encrypt", "decrypt"]
  );

  const decrypted = await crypto.subtle.decrypt(
    { name: "AES-GCM", iv: iv },
    importedKey,
    data
  );
  return new TextDecoder().decode(decrypted);
};

const SHA1 = async (
  value: string | ArrayBuffer,
  encoding = "utf8"
): Promise<ArrayBuffer> => {
  const bytes = getBytes(value, encoding);
  return await crypto.subtle.digest("SHA-1", bytes);
};

const SHA256 = async (
  value: string | ArrayBuffer,
  encoding = "utf8"
): Promise<ArrayBuffer> => {
  const bytes = getBytes(value, encoding);
  return await crypto.subtle.digest("SHA-256", bytes);
};

const SHA384 = async (
  value: string | ArrayBuffer,
  encoding = "utf8"
): Promise<ArrayBuffer> => {
  const bytes = getBytes(value, encoding);
  return await crypto.subtle.digest("SHA-384", bytes);
};

const SHA512 = async (
  value: string | ArrayBuffer,
  encoding = "utf8"
): Promise<ArrayBuffer> => {
  const bytes = getBytes(value, encoding);
  return await crypto.subtle.digest("SHA-512", bytes);
};

const MD5 = async (
  value: string | ArrayBuffer,
  encoding = "utf8"
): Promise<ArrayBuffer> => {
  const bytes = getBytes(value, encoding);
  return md5.arrayBuffer(bytes);
};

const getBytes = (value: string | ArrayBuffer, encoding): ArrayBuffer => {
  let bytes;
  if (typeof value === "string") {
    if (encoding === "base64") {
      bytes = base64ToArrayBuffer(value);
    } else {
      const encoder = new TextEncoder();
      bytes = encoder.encode(value);
    }
  } else {
    bytes = value;
  }
  return bytes;
};

const implementation: ICryptoService = {
  encryptAESGCM,
  decryptAESGCM,
  verifyRSASignature,
  generateRSAKeys(): Promise<RSAKeys> {
    throw new Error("Surprise. Not implemented yet :)");
  },
  encryptRSA(publicKeyPEM: string, data: ArrayBuffer): Promise<ArrayBuffer> {
    throw new Error("Surprise. Not implemented yet :)");
  },
  decryptRSA(privateKeyPEM: string, data: ArrayBuffer): Promise<ArrayBuffer> {
    throw new Error("Surprise. Not implemented yet :)");
  },
  signRSA(privateKeyPEM: string, data: ArrayBuffer): Promise<ArrayBuffer> {
    throw new Error("Surprise. Not implemented yet :)");
  },
  SHA1,
  SHA384,
  SHA256,
  SHA512,
  MD5,
};

export default implementation;
