import { HTMLNormalizer, PlainNormalizer } from "@vereign/mime-normalizer";
import { DOM } from "@vereign/dom";

import MailParser, { fixNewLines } from "../MailParser/MailParser";
import CryptoService from "../CryptoService";
import MimeVerificationService from "./index";
import RKAUtility from "../../utils/rka";
import { arrayBufferToBase64 } from "../../utils/common";
import VerificationError from "../VerificationService/VerificationError";

/**
 * @deprecated
 */
const prepareMimeForPartsVerification = async (
  mimeString: string,
  attachmentsHashes: Array<string>,
  senderSystem: string,
  sourceHtmlRabinFingerprint: string,
  sourceHtmlSize: number,
  sourcePlainRabinFingerprint: string,
  sourcePlainSize: number
): Promise<string> => {
  const { htmlPart, plainPart } = getNormalizedMimeParts(
    mimeString,
    senderSystem
  );

  const originalHtml = findSubstring(
    htmlPart,
    sourceHtmlRabinFingerprint,
    sourceHtmlSize
  );
  if (originalHtml === null) {
    throw new VerificationError(
      "Unable to verify MIME. Source HTML part was not found."
    );
  }

  const originalPlain = findSubstring(
    plainPart,
    sourcePlainRabinFingerprint,
    sourcePlainSize
  );
  if (originalPlain === null) {
    throw new VerificationError(
      "Unable to verify MIME. Source Plain part was not found."
    );
  }

  return await MimeVerificationService.preparePartsForSigning([
    originalHtml,
    originalPlain,
  ]);
};

const getNormalizedMimeParts = (
  mimeString: string,
  senderSystem: string
): { htmlPart: string; plainPart: string } => {
  const {
    htmlPart,
    plainPart,
  } = MimeVerificationService.getMimeHtmlAndPlainParts(mimeString);

  const normalizedHtmlPart = normalizeVendorHtml(htmlPart, senderSystem);

  const normalizedPlainPart = normalizePlainPart(plainPart);

  return {
    htmlPart: normalizedHtmlPart,
    plainPart: normalizedPlainPart,
  };
};

const findSubstring = (
  string: string,
  rabinFingerprint: string,
  substringSize: number
): string => {
  if (!string.length) {
    return "";
  }

  const startingIndex = RKAUtility.searchFingerprintInText(
    string,
    rabinFingerprint,
    substringSize
  );

  if (startingIndex === -1) {
    return null;
  }

  return string.slice(startingIndex, startingIndex + substringSize);
};

const getMimeHtmlAndPlainParts = (
  mimeString: string
): {
  htmlPart: string;
  plainPart: string;
} => {
  mimeString = fixNewLines(mimeString);
  const mimeParts = MailParser.parseMIME(mimeString);

  // TODO: check, whether we need to support 2+ html/plain MIME parts.
  const htmlPart = MailParser.getHTML(mimeString, mimeParts);
  const plainPart = MailParser.getPlain(mimeString, mimeParts);

  return {
    htmlPart,
    plainPart,
  };
};

/**
 * Function implementing specification
 * https://code.vereign.com/light/documentation/-/blob/master/Validation.md#partssignaturealgorithm
 *
 * @param parts - array of strings representing parts
 * @param options - hashing options
 */
const preparePartsForSigning = async (
  parts: Array<string>,
  options: {
    hashParts: boolean;
    hashFunction?: (text: string) => Promise<string>;
  } = {
    hashParts: true,
  }
): Promise<string> => {
  if (options.hashParts) {
    let hashFunction = options.hashFunction;

    if (!hashFunction) {
      hashFunction = async (text: string) => {
        return arrayBufferToBase64(await CryptoService.SHA256(text));
      };
    }

    parts = await Promise.all(
      parts.map(async (part) => {
        return await hashFunction(part);
      })
    );
  }

  return parts.sort().join("\n");
};

const normalizeVendorHtml = (htmlString: string, vendor: string): string => {
  let document;
  if (typeof DOMParser !== "undefined") {
    const parser = new DOMParser();
    document = parser.parseFromString(htmlString, "text/html");
  } else {
    const { window } = new DOM(htmlString);
    document = window.document;
  }

  return HTMLNormalizer.normalizeVendorHtml(document, vendor);
};

const normalizePlainPart = PlainNormalizer.normalizePlain;

export default {
  normalizeVendorHtml,
  getMimeHtmlAndPlainParts,
  preparePartsForSigning,
  prepareMimeForPartsVerification,
  getNormalizedMimeParts,
  normalizePlainPart,
};
